#!/usr/bin/env python

#
# TODO:
#
# Prefix & surfix
# Length of Line
# Change input/output options
# Encoding? (' ' -> %20)
# EOL (Windows vs Linux)

import sys
import os
from optparse import OptionParser


#########################
# Start ExeWriter class #
#########################

class ExeWriter:
    # Initialization object configuration
    def __init__(self, exe_file, bat_file, posh_file, output_dir):
        self.exe_file = exe_file
        self.exe_filename = ''
        self.bat_file = bat_file
        self.bat_short_file = ''
        self.posh_file = posh_file
        self.exe_bin = ''
        self.posh_hex = ''
        self.bat_hex = ''
        self.byte_count = 0
        self.bin_size = 0
        self.output_dir = output_dir

        # Make sure to include a trailing space for the output directory name
        if self.output_dir[-1] != '/':
            self.output_dir += '/'

        # Extract the input filename from the input path
        self.exe_filename = os.path.basename(self.exe_file)

    # Make sure the input file exists
    def check_exe(self):
        if not os.path.isfile(self.exe_file):
            error_exit("The input file was not found  (%s)" % (self.exe_file))

    # Make sure the output directory exists
    def check_output_dir(self):
        if not os.path.isdir(self.output_dir):
            error_exit("The output directory was not found (%s)" % self.output_dir)

    # Make sure the binary size <= 64k when using bat files (limitation with debug.exe)
    def check_bat_size(self):
        if self.bin_size > 65536:
            error_exit("The input file must be under 64k (%d/65536) when using BAT output." % (self.bin_size))

    # Get the contents of the input file
    def read_bin_file(self):
        # Open the input file
        with open(self.exe_file, "rb") as f:
            # Loop forever
            while 1:
                # Read file
                try:
                    byte = f.read(1)
                except:
                    error_exit("A problem occurred while reading the input file (%s)" % self.exe_file)

                # The last byte will have "0" for its length. Break the loop.
                if len(byte) == 0:
                    break

                # Add the read byte into the byte string
                self.exe_bin += byte

        # Set the size of the input file
        self.bin_size = os.path.getsize(self.exe_file)

    # Get the contents of STDIN input
    def read_bin_stdin(self):
        # Read from STDIN
        f = ''
        try:
            f = sys.stdin.read()
        except:
            error_exit("A problem occurred while reading STDIN")

        # Get the length of data read
        stdin_bytes = len(f)

        # Did something go wrong?
        if stdin_bytes == 0:
            error_exit("Zero bytes read from STDIN)") % stdin_bytes

        # Set the size from STDIN
        self.bin_size = stdin_length

        # Add the read byte into the byte string
        for byte in f:
            self.exe_bin += byte

    # Convert binary data to a bat file
    def bin_to_bat_hex(self):
        # Counter for how many bytes have been looped
        byte_count = 0

        # debug.exe has a limitation when renaming files > 8 characters (8.3 filename).
        self.bat_short_file = os.path.splitext(self.bat_file)[0][:8]

        # Loop through binary bytes
        for byte in self.exe_bin:
            # Every 128 bytes. New line.
            if ((byte_count % 128) == 0):
                # Is this anything but the first line?
                if byte_count != 0:
                    # End the line
                    self.bat_hex += ' >>%s.hex\n' % self.bat_short_file
                # Start a new line
                self.bat_hex += 'echo e %s >>%s.hex\necho' % (
                    '{:04x}'.format(byte_count + 256), self.bat_short_file)

            # Add ASCII hex byte
            self.bat_hex += ' {:02x}'.format(ord(byte))

            # Increment byte counter
            byte_count += 1

        # Add the last output line
        self.bat_hex += ' >>%s.hex\n' % self.bat_short_file

        # Save byte counter (debug.exe needs it)
        self.byte_count = byte_count

        # Convert binary data to a PoSh file

    def bin_to_posh_hex(self):
        # Counter for how many bytes have been looped
        byte_count = 0

        # Loop through binary bytes
        for byte in self.exe_bin:
            # Every 128 bytes. New line.
            if ((byte_count % 128) == 0):
                # Is this anything but the first line?
                if byte_count != 0:
                    self.posh_hex += '>'
                # End & start a new line
                self.posh_hex += '"> %s.txt\n<NUL set /p ="' % self.posh_file

            # Append ASCII hex byte
            self.posh_hex += '{:02x}'.format(ord(byte))

            # Increment byte counter
            byte_count += 1

        # Add the last output line
        self.posh_hex += '">> %s.txt' % self.posh_file

    # Write output
    def write_file(self, filepath, contents):
        try:
            f = open(filepath, 'w')
            f.write(contents)
            f.close
            success_msg("Successfully wrote: %s" % filepath)
        except:
            error_msg("A problem occurred while writing (%s)" % filepath)

    # Write resulting bat file
    def save_bat(self):
        # Check size due to limitation of debug,exe (<= 64k)
        self.check_bat_size()

        # Create bat file!
        output = 'echo n %s.dll >%s.hex\n' % (self.bat_short_file, self.bat_short_file)
        output += self.bat_hex
        output += 'echo r cx >>%s.hex\n' % self.bat_short_file
        output += 'echo %s >>%s.hex\n' % ('{:04x}'.format(self.byte_count), self.bat_short_file)
        output += 'echo w >>%s.hex\n' % self.bat_short_file
        output += 'echo q >>%s.hex\n' % self.bat_short_file
        output += 'debug<%s.hex\n' % self.bat_short_file
        output += 'move %s.dll %s\n' % (self.bat_short_file, self.exe_filename)
        output += '%s\n' % self.exe_filename

        # Write file out
        self.write_file(self.output_dir + self.bat_file, output)

    # Write resulting PoSh file
    def save_posh(self):
        # Create PoSh file!
        output = '<NUL set /p ="'
        output += self.posh_hex
        output += "\npowershell $hs=Get-Content -readcount 0 -path './%s.txt';\n" % self.posh_file
        output += "$hs = $hs[0]; $c = $hs.length; $bc = $c/2; $bs = New-Object byte[] $bc;\n"
        output += "$b = $null; $x=0; for ( $i = 0; $i -le $c-1; $i+=2 )\n"
        output += "{ $bs[$x] = [byte]::Parse($hs.Substring($i,2),\n"
        output += "[System.Globalization.NumberStyles]::HexNumber); $x+=1 }\n"
        output += "./%s\n" % self.exe_filename

        # Write file out
        self.write_file(self.output_dir + self.posh_file, output)

    # Main action
    def run(self):
        # If there is a EXE input, check its valid
        if self.exe_file != None:
            self.check_exe()

        # Check the output folder
        self.check_output_dir()

        # Read binary data (file or STDIN)
        if self.exe_file != None:
            self.read_bin_file()
        else:
            self.notification_msg("Reading from STDIN")
            print ""
            self.read_bin_stdin()

        # Make bat file
        if self.bat_file != None:
            self.bin_to_bat_hex()
            self.save_bat()

        # Make PoSh file
        if self.posh_file != None:
            self.bin_to_posh_hex()
            self.save_posh()


#######################
# End ExeWriter class #
#######################

###################
# Functions start #
###################

# Standard error message and exit
def error_exit(msg):
    error_msg(msg)
    sys.exit(1)


# Standard error message
def error_msg(msg):
    sys.stderr.write("\033[01;31m[!]\033[00m ERROR: %s\n" % msg)


# Standard success message
def success_msg(msg):
    print("\033[01;32m[+]\033[00m %s" % msg)


# Standard notification message
def notification_msg(msg):
    print("\033[01;33m[i]\033[00m %s" % msg)


# Banner Information
def banner_msg(msg):
    print("\033[01;34m[*]\033[00m %s" % msg)


#################
# Functions End #
#################

################
# Main Program #
################

# Only run if we are being used as stand-alone script
if __name__ == "__main__":
    # Display banner
    banner_msg("exe2hex v1.0")
    print ""

    # Configure command-line option parsing
    parser = OptionParser()
    parser.add_option("-x", dest="exe",
                      help="EXE binary file to convert", metavar="EXE")
    parser.add_option("-s", dest="stdin",
                      help="Read from STDIN", metavar="STDIN")
    parser.add_option("-b", dest="bat",
                      help="BAT output file name (DEBUG.exe method)", metavar="BAT")
    parser.add_option("-p", dest="posh",
                      help="PoSh output file name (Powershell method)", metavar="POSH")
    parser.add_option("-o", dest="output",
                      help="OUTPUT directory for generated files", metavar="OUTPUT")

    # Store command-line options and arguments in variables
    (options, args) = parser.parse_args()
    exe = options.exe
    bat = options.bat
    posh = options.posh
    out_dir = options.output
    stdin = options.stdin

    # Is there any arguments?
    if len(args) < 3:
        banner_msg("Converts executable binary files into ascii text output")
        banner_msg("Able to outputs using bat/debug.exe method or PoSh (POwerSHell) format(s)")
        print ""
        banner_msg("Specify binary input with -s or -x")
        banner_msg("Specify output with -b and/or -p")
        banner_msg("More details are available with -h/--help")
        print ""

    # Any input methods?
    if exe == None and stdin == None:
        parser.print_help()

        error_exit("Missing executable file or STDIN must be specified")
        parser.print_help()

    # Too many input methods?
    if exe != None and stdin != None:
        error_exit("Cannot use both file and stdin")
        parser.print_help()

    # Any output methods?
    if bat == None and posh == None:
        error_exit("A bat and/or PoSh output file must be specified")
        parser.print_help()

    # Any output methods?
    if out_dir == None:
        notification_msg("Notice: Output directory not specified, saving file(s) to PWD")
        out_dir = "./"
        print os.getcwd()

    # Read file information
    e = ExeWriter(exe, bat, posh, out_dir)

    # GO!
    e.run()
